unit Sprite;

(*********************************************
TSprite->TObject

The base class for all sprites.  This is an abstract
class, because of the Render method that MUST be
overriden in derived classes.
*********************************************)

interface

uses
  SysUtils, WinTypes, WinProcs, Classes, Graphics, Controls, WinGPnt;

type

  TMotionType = ( mtStopAtDest, mtContinuous );

  TSprite = class( TObject )
  private
     nCycle: word;
     nDX, nDY: integer;
     nIncX, nIncY: integer;
     nError: integer;
     nInc: byte;
     FPri: integer;
     bRegistered: boolean;
     FMotion: TMotionType;
     procedure DoRegister;
  protected
     w: TWinGPaintBox;
     ptDestination: TPoint;
     nSpeed: byte;
     procedure SetDestination( const pt: TPoint );
     procedure SetSpeed( const n: byte );
     procedure SetPriority( const n: integer );
  public
     Tag: integer;              { it's nice to have! }
     bKill: boolean;
     ptPosition: TPoint;        { Current logical position }
     ptPhysical: TPoint;        { Logical location - offsets }
     constructor Create;
     procedure RegisterSprite( winGPnt: TWinGPaintBox );
     procedure UnRegisterSprite;
     procedure Kill; dynamic;            { A safe way to UnRegister }
     procedure Move; dynamic;
     procedure Render( winGPnt: TWinGPaintBox ); dynamic; abstract;
     function FudgedDistance( s: TSprite ): word;
     property Destination: TPoint read ptDestination write SetDestination;
     property Position: TPoint read ptPosition write ptPosition;
     property Priority: integer read FPri write SetPriority;
     property Speed: byte read nSpeed write SetSpeed;
     property MotionType: TMotionType read FMotion write FMotion;
  end;

implementation

constructor TSprite.Create;
begin
  w := nil;
  Priority := 1;
  ptPosition := Point( 0, 0 );
  SetSpeed( 20 );
  SetDestination( Point( 0, 0 ) );
  MotionType := mtStopAtDest;
  bRegistered := FALSE;
end;

procedure TSprite.RegisterSprite( winGPnt: TWinGPaintBox );
begin
  w := winGPnt;
  if not bRegistered then
     DoRegister;
end;

procedure TSprite.UnRegisterSprite;
var
  i: integer;
begin
  if ( w <> nil ) and bRegistered then
     begin
        i := w.lstSprites.IndexOf( self );
        if i >= 0 then
           begin
              w.lstSprites.Delete( i );
              Free;
           end;
     end;
end;

(*********************************************
if a sprite's priority is changed, it must be
physically moved in the sprite list.
*********************************************)
procedure TSprite.SetPriority( const n: integer );
var
  i: integer;
begin
  FPri := n;
  if bRegistered then
     begin
{ Important to traverse list backwards when changing priority }
        for i := w.lstSprites.Count - 1 downto 0 do
           if TSprite( w.lstSprites[i] ) = self then
              begin
                 w.lstSprites.Delete( i );
                 Break;
              end;
        DoRegister;
     end;
end;

procedure TSprite.Move;
var
  bMoveX, bMoveY: boolean;
begin
{ See if the sprite should be killed }
  if bKill then
     begin
        UnregisterSprite;
        Exit;
     end;

{ Handle wrapping if it's enabled }
  if w.WrapHorizontal then
     begin
        if ptPosition.X < 0 then
           ptPosition.X := w.PhysicalWidth;
        if ptPosition.X > w.PhysicalWidth then
           ptPosition.X := 0;
     end;
  if w.WrapVertical then
     begin
        if ptPosition.Y < 0 then
           ptPosition.Y := w.PhysicalHeight;
        if ptPosition.Y > w.PhysicalHeight then
           ptPosition.Y := 0;
     end;

{ Adjust physical position based on offset into logical space }
  ptPhysical := ptPosition;
  Dec( ptPhysical.X, w.OffsetX );
  Dec( ptPhysical.Y, w.OffsetY );

  bMoveX := TRUE;
  bMoveY := TRUE;

{ Check to see if sprite has reached its destination }
  if FMotion = mtStopAtDest then
     begin
        if nIncX > 0 then
           begin
              if ptPosition.X >= ptDestination.X then
                 bMoveX := FALSE;
           end
        else
           begin
              if ptPosition.X <= ptDestination.X then
                 bMoveX := FALSE;
           end;
        if nIncY > 0 then
           begin
              if ptPosition.Y >= ptDestination.Y then
                 bMoveY := FALSE;
           end
        else
           begin
              if ptPosition.Y <= ptDestination.Y then
                 bMoveY := FALSE;
           end;
     end;

  if bMoveX or bMoveY then
     begin
        Inc( nCycle );
        if nCycle >= nSpeed then
           begin
              nCycle := 0;
              if nDX > nDY then
                 begin
                    Inc( nError, nDY );
                    if nError > nDX then
                       begin
                          Dec( nError, nDX );
                          if bMoveY then
                             Inc( ptPosition.Y, nIncY );
                       end;
                    if bMoveX then
                       Inc( ptPosition.X, nIncX );
                 end
              else
                 begin
                    Inc( nError, nDX );
                    if nError > 0 then
                       begin
                          Dec( nError, nDY );
                          if bMoveX then
                             Inc( ptPosition.X, nIncX );
                       end;
                    if bMoveY then
                       Inc( ptPosition.Y, nIncY );
                 end;
           end;
     end;
end;

procedure TSprite.SetDestination( const pt: TPoint );
begin
  nError := 0;
  ptDestination := pt;
  nDX := ptDestination.X - ptPosition.X;
  nDY := ptDestination.Y - ptPosition.Y;
  if nDX >= 0 then
     nIncX := nInc
  else
     begin
        nIncX := -nInc;
        nDX := -nDX;
     end;
  if nDY >= 0 then
     nIncY := nInc
  else
     begin
        nIncY := -nInc;
        nDY := -nDY;
     end;
end;

procedure TSprite.SetSpeed( const n: byte );
begin
  nSpeed := 0;
  if n < 7 then
     nInc := 3
  else if n < 15 then
     nInc := 2
  else
     begin
        nInc := 1;
        nSpeed := n - 15;
     end;
end;

(*********************************************
Register the sprite, without checked to see
if it is alread registered.
*********************************************)
procedure TSprite.DoRegister;
var
  nIndex, i: integer;
begin
  bRegistered := TRUE;
  nIndex := -1;
  for i := 0 to w.lstSprites.Count - 1 do
     if TSprite( w.lstSprites[i] ).FPri >= FPri then
        begin
           nIndex := i;
           Break;
        end;
  if nIndex = -1 then
     w.lstSprites.Add( self )
  else
     w.lstSprites.Insert( nIndex, self );
end;

(*********************************************
Determine a "fudged" distance by simply adding
the absolute values of the positions.  Faster
than executing the correct distance formula.
*********************************************)
function TSprite.FudgedDistance( s: TSprite ): word;
begin
  Result := Abs( ptPosition.X - s.ptPosition.X ) + Abs( ptPosition.Y - s.ptPosition.Y );
end;

(*********************************************
Calling Kill will set the bKill flag, and the
sprite will unregister itself in its next
execute of its Move method.
*********************************************)
procedure TSprite.Kill;
begin
  bKill := TRUE;
end;

end.
